# DOWNLOADED AND CACHED DATA - DO NOT EDIT - see README for more info

[('all', False),
 ('alpha', True),
 ('amd64', False),
 ('arm64', False),
 ('armel', False),
 ('armhf', False),
 ('hppa', True),
 ('hurd-amd64', True),
 ('hurd-i386', True),
 ('i386', False),
 ('loong64', True),
 ('m68k', True),
 ('mips64el', False),
 ('mipsel', False),
 ('powerpc', True),
 ('ppc64', True),
 ('ppc64el', False),
 ('riscv64', False),
 ('s390x', False),
 ('sh4', True),
 ('sparc64', True),
 ('x32', True)]
